(******************************************************************************)
(* computer (stack machine)                                                   *)
(* (C) 2010..2021 Steffen Solyga <solyga@gmx.de>                              *)
(******************************************************************************)
IMPLEMENTATION MODULE Comp;

FROM SYSTEM IMPORT ADR, BYTE;
IMPORT Lib, FIO;

CONST
  debugG = FALSE;
  debugX = TRUE;
  out ::= FIO.StandardOutput;
  err ::= FIO.ErrorOutput;
  dbg ::= FIO.ErrorOutput;

  acMax = 32;
  vcMax = 32;     (* TIndex *)
  pcMax = 1024;   (* TAddress *)

TYPE
  TFunction = ( fHlt, fVal, fLod, fSto, fOpr );
  TCmdDesc = RECORD
    CASE fct: TFunction OF
    | fHlt      : (* empty *)
    | fVal      : val: TValue;
    | fLod, fSto: idx: TIndex;
    | fOpr      : opr: TOperation;
    END;
  END;
  TAddress = CARDINAL;

VAR
  (* program *)
  prg: ARRAY [ 0 .. pcMax - 1 ] OF BYTE;
  len: TAddress;
  (* machine state (execution) *)
  args: ARRAY [ 1 .. acMax ] OF TValue;  (* stack *)
  ac: CARDINAL;
  vars: ARRAY [ 0 .. vcMax - 1 ] OF TValue;  (* memory *)
  pos: TAddress;
  (* module state *)
  strs: ARRAY BOOLEAN OF ARRAY [ 0 .. 4 ] OF CHAR;
  firstG, firstX: BOOLEAN;

(***** debugging *****)

PROCEDURE DbgH( gen: BOOLEAN );
BEGIN
  IF gen THEN
    FIO.WrStr( dbg, 'Comp.DebugG:  pos' );
  ELSE
    FIO.WrStr( dbg, 'Comp.DebugX: ac       top-1         top' );
  END;
  FIO.WrLn( dbg );
END DbgH;

PROCEDURE Dbg( s: ARRAY OF CHAR; gen: BOOLEAN );
BEGIN
  IF gen THEN
    IF firstG THEN DbgH( gen ); firstG := FALSE END;
    FIO.WrStr( dbg, 'Comp.DebugG: ' );
    FIO.WrCard( dbg, len, 4 );
  ELSE
    IF firstX THEN DbgH( gen ); firstX := FALSE END;
    FIO.WrStr( dbg, 'Comp.DebugX: ' );
    FIO.WrCard( dbg, ac, 2 );
    FIO.WrChar( dbg, ' ' );
    IF ac = 0 THEN
      FIO.WrStr( dbg, '                       ' );
    ELSIF ac = 1 THEN
      FIO.WrStr( dbg, '            ' );
      FIO.WrInt( dbg, args[ac], 11 );
    ELSE
      FIO.WrInt( dbg, args[ac-1], 11 );
      FIO.WrChar( dbg, ' ' );
      FIO.WrInt( dbg, args[ac], 11 );
    END;
  END;
  FIO.WrStr( dbg, ':  ' );
  FIO.WrStr( dbg, s );
END Dbg;

PROCEDURE DbgV( v: TValue; gen: BOOLEAN );
BEGIN
  Dbg( 'Val', gen );
  FIO.WrStr( dbg, ' ' );
  FIO.WrInt( dbg, v, 0 );
  FIO.WrLn( dbg );
END DbgV;

PROCEDURE DbgI( s: ARRAY OF CHAR; i: TIndex; gen: BOOLEAN );
BEGIN
  Dbg( s, gen );
  FIO.WrStr( dbg, ' ' );
  FIO.WrCard( dbg, i, 1 );
  FIO.WrLn( dbg );
END DbgI;

PROCEDURE DbgO( o: TOperation; gen: BOOLEAN );
BEGIN
  CASE o OF
  | oNul: Dbg( 'Nul', gen );
  | oWrt: Dbg( 'Wrt', gen );
  | oWrB: Dbg( 'WrB', gen );
  | oNeg: Dbg( 'Neg', gen );
  | oAdd: Dbg( 'Add', gen );
  | oSub: Dbg( 'Sub', gen );
  | oMul: Dbg( 'Mul', gen );
  | oDiv: Dbg( 'Div', gen );
  | oEql: Dbg( 'Eql', gen );
  | oNeq: Dbg( 'Neq', gen );
  | oLss: Dbg( 'Lss', gen );
  | oLeq: Dbg( 'Leq', gen );
  | oGtr: Dbg( 'Gtr', gen );
  | oGeq: Dbg( 'Geq', gen );
  | oNot: Dbg( 'Not', gen );
  | oOr : Dbg( 'Or ', gen );
  | oAnd: Dbg( 'And', gen );
  ELSE
    Dbg( '???', gen );
  END;
  FIO.WrLn( dbg );
END DbgO;


PROCEDURE Err( s: ARRAY OF CHAR );
BEGIN
  FIO.WrStr( err, 'Comp: ' );
  FIO.WrStr( err, s );
  FIO.WrStr( err, '.' );
  FIO.WrLn( err );
  HALT;
END Err;

(***** execution *****)

PROCEDURE Read( VAR x: ARRAY OF BYTE ): BOOLEAN;
VAR i: CARDINAL;
BEGIN
  IF pos + SIZE( x ) > len THEN RETURN FALSE END;
  FOR i := 0 TO HIGH( x ) DO x[ i ] := prg[ pos ]; INC( pos ) END;
  RETURN TRUE;
END Read;

PROCEDURE Idx( i: TIndex ): TIndex;
BEGIN
  IF i >= vcMax THEN Err( 'Invalid index' ) END;
  RETURN i;
END Idx;

PROCEDURE Push( x: TValue );
BEGIN
  IF ac >= acMax THEN Err( 'Stack overflow' ) END;
  INC( ac ); args[ ac ] := x;
END Push;

PROCEDURE Pop( VAR x: TValue );
BEGIN
  IF ac <= 0 THEN Err( 'Stack underflow' ) END;
  x := args[ ac ]; DEC( ac );
END Pop;

PROCEDURE True( x: TValue ): BOOLEAN;
BEGIN
  RETURN x # 0;
END True;

PROCEDURE Exec();
VAR
  cmd: TCmdDesc;
  x,y: TValue;
BEGIN
  WHILE pos < len DO
    IF ~ Read( cmd.fct ) THEN RETURN END;
    CASE cmd.fct OF
    | fHlt: (* empty *)
            IF debugX THEN Dbg( 'Hlt', FALSE ) END;
            RETURN;
    | fVal: IF ~ Read( cmd.val ) THEN Err( 'Bug: Incomplete command' ) END;
            IF debugX THEN DbgV( cmd.val, FALSE ) END;
            Push( cmd.val );
    | fLod: IF ~ Read( cmd.idx ) THEN Err( 'Bug: Incomplete command' ) END;
            IF debugX THEN DbgI( 'Lod', cmd.idx, FALSE ) END;
            Push( vars[Idx(cmd.idx)] );
    | fSto: IF ~ Read( cmd.idx ) THEN Err( 'Bug: Incomplete command' ) END;
            IF debugX THEN DbgI( 'Sto', cmd.idx, FALSE ) END;
            Pop( vars[Idx(cmd.idx)] );
    | fOpr: IF ~ Read( cmd.opr ) THEN Err( 'Bug: Incomplete command' ) END;
            IF debugX THEN DbgO( cmd.opr, FALSE ) END;
            CASE cmd.opr OF
            | oWrt: Pop( x ); FIO.WrInt( out, x, 0 ); FIO.WrLn( out );
            | oWrB: Pop( x ); FIO.WrStr( out, strs[True(x)] ); FIO.WrLn( out );
            | oNeg: Pop( x ); Push( -x );
            | oAdd: Pop( y ); Pop( x ); Push( x + y );
            | oSub: Pop( y ); Pop( x ); Push( x - y );
            | oMul: Pop( y ); Pop( x ); Push( x * y );
            | oDiv: Pop( y ); Pop( x );
                    IF y = 0 THEN Err( 'Division by zero' ) END;
                    Push( x DIV y );
            | oEql: Pop( y ); Pop( x ); Push( ORD( x=y ) );
            | oNeq: Pop( y ); Pop( x ); Push( ORD( x#y ) );
            | oLss: Pop( y ); Pop( x ); Push( ORD( x<y ) );
            | oLeq: Pop( y ); Pop( x ); Push( ORD( x<=y ) );
            | oGtr: Pop( y ); Pop( x ); Push( ORD( x>y ) );
            | oGeq: Pop( y ); Pop( x ); Push( ORD( x>=y ) );
            | oNot: Pop( x ); Push( ORD( ~True(x) ) );
            | oOr : Pop( y ); Pop( x ); Push( ORD( True(x) OR True(y) ) );
            | oAnd: Pop( y ); Pop( x ); Push( ORD( True(x) & True(y) ) );
            ELSE
              Err( 'Bug: Invalid operation' );
            END;
    ELSE
      Err( 'Bug: Invalid function' );
    END;
  END;
END Exec;

(***** programming *****)

PROCEDURE Write( x: ARRAY OF BYTE );
VAR i: CARDINAL;
BEGIN
  IF len + SIZE( x ) > pcMax THEN Err( 'Program too large' ) END;
  FOR i := 0 TO HIGH( x ) DO prg[ len ] := x[ i ]; INC( len ) END;
END Write;


PROCEDURE Val( v: TValue );
VAR cmd: TCmdDesc;
BEGIN
  IF debugG THEN DbgV( v, TRUE ) END;
  cmd.fct := fVal; cmd.val := v;
  Write( cmd.fct ); Write( cmd.val );
  Exec();
END Val;

PROCEDURE Lod( i: TIndex );
VAR cmd: TCmdDesc;
BEGIN
  IF debugG THEN DbgI( 'Lod', i, TRUE ) END;
  cmd.fct := fLod; cmd.idx := Idx( i );
  Write( cmd.fct ); Write( cmd.idx );
  Exec();
END Lod;

PROCEDURE Sto( i: TIndex );
VAR cmd: TCmdDesc;
BEGIN
  IF debugG THEN DbgI( 'Sto', i, TRUE ) END;
  cmd.fct := fSto; cmd.idx := Idx( i );
  Write( cmd.fct ); Write( cmd.idx );
  Exec();
END Sto;

PROCEDURE Opr( o: TOperation );
VAR cmd: TCmdDesc;
BEGIN
  IF debugG THEN DbgO( o, TRUE ) END;
  IF o = oNul THEN Err( 'Bug: Invalid operation' ) END;
  cmd.fct := fOpr; cmd.opr := o;
  Write( cmd.fct ); Write( cmd.opr );
  Exec();
END Opr;

BEGIN
  (* program *)
  Lib.Fill( ADR(prg), SIZE(prg), fHlt );
  len := 0;
  (* machine state *)
  Lib.Fill( ADR(args), SIZE(args), 0C );
  ac := 0;
  Lib.Fill( ADR(vars), SIZE(vars), 0C );
  pos := 0;
  (* module state *)
  firstG := TRUE;
  firstX := TRUE;
  strs[ FALSE ] := 'FALSE';
  strs[ TRUE ] := 'TRUE';
END Comp.
