(******************************************************************************)
(* plx parser                                                                 *)
(* (C) 2010..2021 Steffen Solyga <solyga@gmx.de>                              *)
(******************************************************************************)
IMPLEMENTATION MODULE Prsr;

IMPORT FIO;
FROM Scnr IMPORT Put, Read, KeepId, Diff, TSymbol, sym, num, id, pos;
FROM Comp IMPORT TValue, TOperation, Val, Opr, Sto, Lod;

CONST
  err ::= FIO.ErrorOutput;
  cntMax = 32;

TYPE
  TClass = ( cConst, cVar );
  TType = ( tNull, tBool, tInt );
  TObjDesc = RECORD
    id : CARDINAL;
    typ: TType;
    CASE cls: TClass OF
    | cConst: val: TValue;
    | cVar  : idx: CARDINAL;
    END;
  END;

VAR
  objs: ARRAY [ 0 .. cntMax - 1 ] OF TObjDesc;
  cnt: CARDINAL;
  idx: CARDINAL;
  oprs: ARRAY TSymbol OF TOperation;

PROCEDURE Err( s: ARRAY OF CHAR );
BEGIN
  FIO.WrStr( err, 'Prsr (' );
  FIO.WrCard( err, pos.l, 0 );
  FIO.WrChar( err, ',' );
  FIO.WrCard( err, pos.c, 0 );
  FIO.WrStr( err, '): ' );
  FIO.WrStr( err, s );
  FIO.WrStr( err, '.' );
  FIO.WrLn( err );
  HALT;
END Err;

PROCEDURE Find( VAR i: CARDINAL );
BEGIN
  i := 0;
  WHILE i < cnt DO
    IF Diff( objs[i].id, id ) = 0 THEN RETURN END;
    INC( i );
  END;
END Find;

PROCEDURE Expression( VAR t: TType );

  PROCEDURE SimpleExpr( VAR t: TType );

    PROCEDURE Factor( VAR t: TType );
    VAR i: CARDINAL; s: TSymbol;
    BEGIN
      IF sym = sNum THEN
        IF num > MAX( TValue ) THEN Err( 'Number too large' ) END;
        Val( TValue(num) ); Read(); t := tInt;
      ELSIF sym = sId THEN
        Find( i );
        IF i >= cnt THEN Err( 'Identifier undeclared' ) END;
        t := objs[ i ].typ;
        IF objs[ i ].cls = cConst THEN
          Val( objs[i].val );
        ELSE
          IF t = tNull THEN Err( 'Variable not initialised' ) END;
          Lod( objs[i].idx );
        END;
        Read();
      ELSIF sym = sLPar THEN
        Read(); Expression( t );
        IF sym = sRPar THEN Read() ELSE Err( 'Expected operator or ")"' ) END;
      ELSIF sym = sNot THEN
        s := sym; Read(); Factor( t );
        IF t # tBool THEN Err( 'Expected boolean factor' ) END;
        Opr( oprs[s] );
      ELSE
        Err( 'Expected factor' );
      END;
    END Factor;

    PROCEDURE Term( VAR t: TType );
    VAR s: TSymbol; t1: TType;
    BEGIN
      Factor( t );
      WHILE ( sym = sMul ) OR ( sym = sDiv ) OR ( sym = sAnd ) DO
        IF ( sym = sAnd ) # ( t = tBool ) THEN Err( 'Wrong operator type' ) END;
        s := sym; Read(); Factor( t1 );
        IF t # t1 THEN Err( 'Wrong argument type' ) END;
        Opr( oprs[s] );
      END;
    END Term;

  VAR s: TSymbol; t1: TType;
  BEGIN
    IF ( sym = sPlus ) OR ( sym = sMinus ) THEN
      s := sym; Read(); Term( t );
      IF t # tInt THEN Err( 'Expected integer term' ) END;
      IF s = sMinus THEN Opr( oNeg ) END;
    ELSE
      Term( t );
    END;
    WHILE ( sym = sPlus ) OR ( sym = sMinus ) OR ( sym = sOr ) DO
      IF ( sym = sOr ) # ( t = tBool ) THEN Err( 'Wrong operator type' ) END;
      s := sym; Read(); Term( t1 );
      IF t # t1 THEN Err( 'Wrong argument type' ) END;
      Opr( oprs[s] );
    END;
  END SimpleExpr;

VAR s: TSymbol; t1: TType;
BEGIN
  SimpleExpr( t );
  IF ( sym = sEql ) OR ( sym = sNeq ) OR ( sym = sLss ) OR
     ( sym = sLeq ) OR ( sym = sGtr ) OR ( sym = sGeq ) THEN
    s := sym; Read(); SimpleExpr( t1 );
    IF t # t1 THEN Err( 'Wrong argument type' ) END;
    Opr( oprs[s] );
    t := tBool;
  END;
END Expression;

PROCEDURE Statement();
VAR oi: CARDINAL; t: TType;
BEGIN
  IF sym = sId THEN
    Find( oi ); Read();
    IF sym = sEql THEN  (* const declaration *)
      IF oi < cnt THEN Err( 'Identifier redeclared' ) END;
      IF cnt >= cntMax THEN Err( 'Too many identifiers' ) END;
      objs[ cnt ].id  := id;
      objs[ cnt ].cls := cConst;
      KeepId(); Read();
      IF sym = sNum THEN
        IF num > MAX( TValue ) THEN Err( 'Number too large' ) END;
        objs[ cnt ].typ := tInt;
        objs[ cnt ].val := TValue( num );
      ELSIF sym = sId THEN
        Find( oi );
        IF oi >= cnt THEN Err( 'Identifier undeclared' ) END;
        IF objs[ oi ].cls # cConst THEN Err( 'Not a constant' ) END;
        objs[ cnt ].typ := objs[ oi ].typ;
        objs[ cnt ].val := objs[ oi ].val;
      ELSE
        Err( 'Expected number or identifier' );
      END;
      INC( cnt ); Read();
    ELSIF sym = sBec THEN  (* assignment *)
      IF oi >= cnt THEN  (* var declaration *)
        IF cnt >= cntMax THEN Err( 'Too many identifiers' ) END;
        objs[ cnt ].id  := id;
        objs[ cnt ].cls := cVar;
        objs[ cnt ].typ := tNull;
        objs[ cnt ].idx := idx;
        KeepId(); INC( cnt ); INC( idx );
      END;
      IF objs[ oi ].cls # cVar THEN Err( 'Not a variable' ) END;
      Read(); Expression( t ); Sto( objs[oi].idx );
      objs[ oi ].typ := t;
    ELSE
      Err( 'Expected "=" or ":="' );
    END;
  ELSIF sym = sExcl THEN   (* expression output *)
    Read(); Expression( t );
    CASE t OF
    | tInt : Opr( oWrt );
    | tBool: Opr( oWrB );
    ELSE
      Err( 'Bug: Invalid type' );
    END;
  END;
END Statement;

PROCEDURE Run(); (* Program() *)
BEGIN
  Read(); Statement();
  WHILE sym = sSemi DO Read(); Statement() END;
  IF sym # sDot THEN Err( 'Expected ";" or "."' ) END;
END Run;

BEGIN
  cnt := 0;
  idx := 0;
  Put( 'FALSE' );
  objs[ cnt ].id  := id;
  objs[ cnt ].typ := tBool;
  objs[ cnt ].cls := cConst;
  objs[ cnt ].val := 0;
  KeepId(); INC( cnt );
  Put( 'TRUE' );
  objs[ cnt ].id  := id;
  objs[ cnt ].typ := tBool;
  objs[ cnt ].cls := cConst;
  objs[ cnt ].val := 1;
  KeepId(); INC( cnt );
  oprs[ sNull  ] := oNul;
  oprs[ sNum   ] := oNul;
  oprs[ sId    ] := oNul;
  oprs[ sExcl  ] := oWrt;
  oprs[ sEql   ] := oEql;
  oprs[ sBec   ] := oNul;
  oprs[ sLPar  ] := oNul;
  oprs[ sMul   ] := oMul;
  oprs[ sDiv   ] := oDiv;
  oprs[ sPlus  ] := oAdd;
  oprs[ sMinus ] := oSub;
  oprs[ sNot   ] := oNot;
  oprs[ sOr    ] := oOr;
  oprs[ sAnd   ] := oAnd;
  oprs[ sNeq   ] := oNeq;
  oprs[ sLss   ] := oLss;
  oprs[ sLeq   ] := oLeq;
  oprs[ sGtr   ] := oGtr;
  oprs[ sGeq   ] := oGeq;
  oprs[ sRPar  ] := oNul;
  oprs[ sSemi  ] := oNul;
  oprs[ sDot   ] := oNul;
  oprs[ sEOF   ] := oNul;
END Prsr.
