(******************************************************************************)
(* plx scanner                                                                *)
(* (C) 2010..2021 Steffen Solyga <solyga@gmx.de>                              *)
(******************************************************************************)
IMPLEMENTATION MODULE Scnr;

IMPORT FIO;

CONST
  bufLen = 1024;
  err ::= FIO.ErrorOutput;
  cEOF = 32C;    (* = 1AH = 26 = ASCII.sub (TS) *)
  cLF = 12C;

TYPE
  TBufIdx = [ 0 .. bufLen - 1 ];
  TBufDesc = ARRAY TBufIdx OF CHAR;

VAR
  chr: CHAR;
  chp: TPosDesc;
  buf: TBufDesc;
  id1: CARDINAL;   (* write index *)

PROCEDURE Err( s: ARRAY OF CHAR );
BEGIN
  FIO.WrStr( err, 'Scnr (' );
  FIO.WrCard( err, pos.l, 0 );
  FIO.WrChar( err, ',' );
  FIO.WrCard( err, pos.c, 0 );
  FIO.WrStr( err, '): ' );
  FIO.WrStr( err, s );
  FIO.WrStr( err, '.' );
  FIO.WrLn( err );
  HALT;
END Err;

PROCEDURE Get();
BEGIN
  IF chr = cEOF THEN RETURN END;
  IF chr = cLF THEN              (* UNIX and DOS text ok *)
    INC( chp.l ); chp.c := 1;
  ELSE
    INC( chp.c );
  END;
  IF FIO.RdBin( src, chr, 1 ) = 0 THEN chr := cEOF END;
END Get;

PROCEDURE Alpha(): BOOLEAN;
BEGIN
  RETURN ( '0' <=     chr  ) & (     chr  <= '9' ) OR
         ( 'A' <= CAP(chr) ) & ( CAP(chr) <= 'Z' );
END Alpha;

PROCEDURE Number();
CONST b = 10;
VAR d: TNumber;
BEGIN
  sym := sNum;
  num := 0;
  REPEAT
    d := ORD( chr ) - ORD( '0' );
    IF num <= ( MAX(TNumber) - d ) DIV b THEN
      num := b * num + d;
    ELSE
      Err( 'Number too large' );
    END;
    Get();
  UNTIL ~ Alpha();
END Number;

PROCEDURE Id();
BEGIN
  sym := sId;
  id1 := id + 1;
  REPEAT
    IF id1 >= bufLen THEN Err( 'Buffer full' ) END;
    buf[ id1 ] := chr; INC( id1 ); Get();
  UNTIL ~ Alpha();
  buf[ id ] := CHR( id1 - id );
END Id;

PROCEDURE KeepId();
BEGIN
  id := id1;
END KeepId;

PROCEDURE Put( s: ARRAY OF CHAR );
VAR i: CARDINAL;
BEGIN
  id1 := id + 1;
  i := 0;
  REPEAT
    IF id1 >= bufLen THEN Err( 'Buffer full' ) END;
    buf[ id1 ] := s[ i ]; INC( i ); INC( id1 );
  UNTIL i >= HIGH( s );
  buf[ id ] := CHR( id1 - id );
END Put;

PROCEDURE Read();

  PROCEDURE Comment();
  VAR lvl: CARDINAL;
  BEGIN
    Get(); lvl := 1;
    WHILE lvl > 0 DO
      IF chr = cEOF THEN chp := pos; Err( 'Unterminated comment (start pos)' ) END;
      CASE chr OF
      | '(' : Get(); IF chr = '*' THEN INC( lvl ); Get() END;
      | '*' : Get(); IF chr = ')' THEN DEC( lvl ); Get() END;
      ELSE
        Get();
      END;
    END;
  END Comment;

BEGIN
  LOOP (* ignore control characters except EOF *)
    IF chr <= ' ' THEN
      IF chr = cEOF THEN EXIT END;
      Get();
    ELSIF chr >= 177C THEN
      Get();
    ELSE
      EXIT;
    END;
  END;
  pos := chp;
  CASE chr OF
  | cEOF    : sym := sEOF;
  | '!'     : sym := sExcl; Get();
  | '#'     : sym := sNeq; Get();
  | '&'     : sym := sAnd; Get();
  | '('     : Get();
              IF chr = '*' THEN Comment(); Read() ELSE sym := sLPar END;
  | ')'     : sym := sRPar; Get();
  | '*'     : sym := sMul; Get();
  | '+'     : sym := sPlus; Get();
  | '-'     : sym := sMinus; Get();
  | '.'     : sym := sDot; Get();
  | '/'     : sym := sDiv; Get();
  | '0'..'9': Number();
  | ':'     : Get(); IF chr # '=' THEN Err( 'Expected "="' ) END;
              sym := sBec; Get();
  | ';'     : sym := sSemi; Get();
  | '<'     : Get();
              CASE chr OF
              | '=': sym := sLeq; Get();
              | '>': sym := sNeq; Get();
              ELSE
                sym := sLss;
              END;
  | '='     : sym := sEql; Get();
  | '>'     : Get();
              IF chr = '=' THEN sym := sGeq; Get() ELSE sym := sGtr END;
  | 'A'..'Z': Id();
  | 'a'..'z': Id();
  | '|'     : sym := sOr; Get();
  | '~'     : sym := sNot; Get();
  ELSE
    Err( 'Invalid character' );
  END;
END Read;

PROCEDURE Diff( i, j: CARDINAL ): INTEGER;
VAR l: CARDINAL;
BEGIN
  l := ORD( buf[i] );  (* always >0 *)
  WHILE l > 0 DO
    IF buf[i] # buf[j] THEN
      RETURN INTEGER( ORD(buf[i]) ) - INTEGER( ORD(buf[j]) );
    END;
    INC( i ); INC( j ); DEC( l );
  END;
  RETURN 0;
END Diff;

BEGIN
  (* local *)
  chr := cLF;
  chp.l := 0;
  chp.c := 0;
  buf[ 0 ] := 0C;
  id1 := 0;
  (* exported *)
  src := FIO.StandardInput;
  sym := sNull;
  num := 0;
  id  := 0;
  pos := chp;
END Scnr.
